@@ -28,25 +28,20 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef SHADER_GLES3_H
-#define SHADER_GLES3_H
+#pragma once
 
 #include "core/math/projection.h"
 #include "core/os/mutex.h"
 #include "core/string/string_builder.h"
 #include "core/templates/hash_map.h"
 #include "core/templates/local_vector.h"
-#include "core/templates/rb_map.h"
 #include "core/templates/rid_owner.h"
-#include "core/variant/variant.h"
 #include "servers/rendering_server.h"
 
 #ifdef GLES3_ENABLED
 
 #include "platform_gl.h"
 
-#include <stdio.h>
-
 class ShaderGLES3 {
 public:
 	struct TextureUniformData {
@@ -107,7 +102,7 @@ private:
 			}
 		};
 
-		LocalVector<OAHashMap<uint64_t, Specialization>> variants;
+		LocalVector<AHashMap<uint64_t, Specialization>> variants;
 	};
 
 	Mutex variant_set_mutex;
@@ -193,29 +188,29 @@ protected:
 		Version *version = version_owner.get_or_null(p_version);
 		ERR_FAIL_NULL_V(version, false);
 
-		if (version->variants.size() == 0) {
+		if (version->variants.is_empty()) {
 			_initialize_version(version); //may lack initialization
 		}
 
-		Version::Specialization *spec = version->variants[p_variant].lookup_ptr(p_specialization);
+		Version::Specialization *spec = version->variants[p_variant].getptr(p_specialization);
 		if (!spec) {
 			if (false) {
 				// Queue load this specialization and use defaults in the meantime (TODO)
 
-				spec = version->variants[p_variant].lookup_ptr(specialization_default_mask);
+				spec = version->variants[p_variant].getptr(specialization_default_mask);
 			} else {
 				// Compile on the spot
 				Version::Specialization s;
 				_compile_specialization(s, p_variant, version, p_specialization);
 				version->variants[p_variant].insert(p_specialization, s);
-				spec = version->variants[p_variant].lookup_ptr(p_specialization);
+				spec = version->variants[p_variant].getptr(p_specialization);
 				if (shader_cache_dir_valid) {
 					_save_to_cache(version);
 				}
 			}
 		} else if (spec->build_queued) {
 			// Still queued, wait
-			spec = version->variants[p_variant].lookup_ptr(specialization_default_mask);
+			spec = version->variants[p_variant].getptr(specialization_default_mask);
 		}
 
 		if (!spec || !spec->ok) {
@@ -233,7 +228,7 @@ protected:
 		Version *version = version_owner.get_or_null(p_version);
 		ERR_FAIL_NULL_V(version, -1);
 		ERR_FAIL_INDEX_V(p_variant, int(version->variants.size()), -1);
-		Version::Specialization *spec = version->variants[p_variant].lookup_ptr(p_specialization);
+		Version::Specialization *spec = version->variants[p_variant].getptr(p_specialization);
 		ERR_FAIL_NULL_V(spec, -1);
 		ERR_FAIL_INDEX_V(p_which, int(spec->uniform_location.size()), -1);
 		return spec->uniform_location[p_which];
@@ -262,5 +257,3 @@ public:
 };
 
 #endif // GLES3_ENABLED
-
-#endif // SHADER_GLES3_H
