@@ -28,8 +28,7 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef RANGE_H
-#define RANGE_H
+#pragma once
 
 #include "scene/gui/control.h"
 
@@ -47,7 +46,7 @@ class Range : public Control {
 		bool allow_lesser = false;
 		HashSet<Range *> owners;
 		void emit_value_changed();
-		void emit_changed(const char *p_what = "");
+		void emit_changed();
 		void redraw_owners();
 	};
 
@@ -59,15 +58,22 @@ class Range : public Control {
 	void _share(Node *p_range);
 
 	void _value_changed_notify();
-	void _changed_notify(const char *p_what = "");
+	void _changed_notify();
 	void _set_value_no_signal(double p_val);
 
 protected:
+	static double _snapped_r128(double p_value, double p_step);
+	double _calc_value(double p_val, double p_step) const;
 	virtual void _value_changed(double p_value);
-	void _notify_shared_value_changed() { shared->emit_value_changed(); };
+	void _notify_shared_value_changed() { shared->emit_value_changed(); }
+	void _notification(int p_what);
 
 	static void _bind_methods();
 
+	void _accessibility_action_inc(const Variant &p_data);
+	void _accessibility_action_dec(const Variant &p_data);
+	void _accessibility_action_set_value(const Variant &p_data);
+
 	bool _rounded_values = false;
 
 	GDVIRTUAL1(_value_changed, double)
@@ -108,5 +114,3 @@ public:
 	Range();
 	~Range();
 };
-
-#endif // RANGE_H
