@@ -28,18 +28,25 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef SPLIT_CONTAINER_H
-#define SPLIT_CONTAINER_H
+#pragma once
 
 #include "scene/gui/container.h"
 
+class TextureRect;
+
 class SplitContainerDragger : public Control {
 	GDCLASS(SplitContainerDragger, Control);
+	friend class SplitContainer;
+	Rect2 split_bar_rect;
 
 protected:
 	void _notification(int p_what);
 	virtual void gui_input(const Ref<InputEvent> &p_event) override;
 
+	void _accessibility_action_inc(const Variant &p_data);
+	void _accessibility_action_dec(const Variant &p_data);
+	void _accessibility_action_set_value(const Variant &p_data);
+
 private:
 	bool dragging = false;
 	int drag_from = 0;
@@ -48,6 +55,8 @@ private:
 
 public:
 	virtual CursorShape get_cursor_shape(const Point2 &p_pos = Point2i()) const override;
+
+	SplitContainerDragger();
 };
 
 class SplitContainer : public Container {
@@ -62,27 +71,47 @@ public:
 	};
 
 private:
+	int show_drag_area = false;
+	int drag_area_margin_begin = 0;
+	int drag_area_margin_end = 0;
+	int drag_area_offset = 0;
 	int split_offset = 0;
-	int middle_sep = 0;
+	int computed_split_offset = 0;
 	bool vertical = false;
 	bool collapsed = false;
 	DraggerVisibility dragger_visibility = DRAGGER_VISIBLE;
+	bool dragging_enabled = true;
 
 	SplitContainerDragger *dragging_area_control = nullptr;
 
+	bool touch_dragger_enabled = false;
+	TextureRect *touch_dragger = nullptr;
+
 	struct ThemeCache {
+		Color touch_dragger_color;
+		Color touch_dragger_pressed_color;
+		Color touch_dragger_hover_color;
 		int separation = 0;
 		int minimum_grab_thickness = 0;
 		bool autohide = false;
+		Ref<Texture2D> touch_dragger_icon;
+		Ref<Texture2D> touch_dragger_icon_h;
+		Ref<Texture2D> touch_dragger_icon_v;
 		Ref<Texture2D> grabber_icon;
 		Ref<Texture2D> grabber_icon_h;
 		Ref<Texture2D> grabber_icon_v;
+		float base_scale = 1.0;
+		Ref<StyleBox> split_bar_background;
 	} theme_cache;
 
 	Ref<Texture2D> _get_grabber_icon() const;
-	void _compute_middle_sep(bool p_clamp);
+	Ref<Texture2D> _get_touch_dragger_icon() const;
+	void _touch_dragger_mouse_exited();
+	void _touch_dragger_gui_input(const Ref<InputEvent> &p_event);
+	void _compute_split_offset(bool p_clamp);
+	int _get_separation() const;
 	void _resort();
-	Control *_get_sortable_child(int p_idx, SortableVisbilityMode p_visibility_mode = SortableVisbilityMode::VISIBLE_IN_TREE) const;
+	Control *_get_sortable_child(int p_idx, SortableVisibilityMode p_visibility_mode = SortableVisibilityMode::VISIBLE_IN_TREE) const;
 
 protected:
 	bool is_fixed = false;
@@ -105,11 +134,31 @@ public:
 	void set_vertical(bool p_vertical);
 	bool is_vertical() const;
 
+	void set_dragging_enabled(bool p_enabled);
+	bool is_dragging_enabled() const;
+
 	virtual Size2 get_minimum_size() const override;
 
 	virtual Vector<int> get_allowed_size_flags_horizontal() const override;
 	virtual Vector<int> get_allowed_size_flags_vertical() const override;
 
+	void set_drag_area_margin_begin(int p_margin);
+	int get_drag_area_margin_begin() const;
+
+	void set_drag_area_margin_end(int p_margin);
+	int get_drag_area_margin_end() const;
+
+	void set_drag_area_offset(int p_offset);
+	int get_drag_area_offset() const;
+
+	void set_show_drag_area_enabled(bool p_enabled);
+	bool is_show_drag_area_enabled() const;
+
+	Control *get_drag_area_control() { return dragging_area_control; }
+
+	void set_touch_dragger_enabled(bool p_enabled);
+	bool is_touch_dragger_enabled() const;
+
 	SplitContainer(bool p_vertical = false);
 };
 
@@ -130,5 +179,3 @@ public:
 	VSplitContainer() :
 			SplitContainer(true) { is_fixed = true; }
 };
-
-#endif // SPLIT_CONTAINER_H
