@@ -28,8 +28,7 @@
 /* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                 */
 /**************************************************************************/
 
-#ifndef NAVIGATION_SERVER_3D_H
-#define NAVIGATION_SERVER_3D_H
+#pragma once
 
 #include "core/object/class_db.h"
 #include "core/templates/rid.h"
@@ -39,12 +38,10 @@
 #include "servers/navigation/navigation_path_query_parameters_3d.h"
 #include "servers/navigation/navigation_path_query_result_3d.h"
 
-/// This server uses the concept of internal mutability.
-/// All the constant functions can be called in multithread because internally
-/// the server takes care to schedule the functions access.
-///
-/// Note: All the `set` functions are commands executed during the `sync` phase,
-/// don't expect that a change is immediately propagated.
+struct NavMeshGeometryParser3D {
+	RID self;
+	Callable callback;
+};
 
 class NavigationServer3D : public Object {
 	GDCLASS(NavigationServer3D, Object);
@@ -55,30 +52,21 @@ protected:
 	static void _bind_methods();
 
 public:
-	/// Thread safe, can be used across many threads.
 	static NavigationServer3D *get_singleton();
 
 	virtual TypedArray<RID> get_maps() const = 0;
 
-	/// Create a new map.
+	/* MAP API */
+
 	virtual RID map_create() = 0;
 
-	/// Set map active.
 	virtual void map_set_active(RID p_map, bool p_active) = 0;
-
-	/// Returns true if the map is active.
 	virtual bool map_is_active(RID p_map) const = 0;
 
-	/// Set the map UP direction.
 	virtual void map_set_up(RID p_map, Vector3 p_up) = 0;
-
-	/// Returns the map UP direction.
 	virtual Vector3 map_get_up(RID p_map) const = 0;
 
-	/// Set the map cell size used to weld the navigation mesh polygons.
 	virtual void map_set_cell_size(RID p_map, real_t p_cell_size) = 0;
-
-	/// Returns the map cell size.
 	virtual real_t map_get_cell_size(RID p_map) const = 0;
 
 	virtual void map_set_cell_height(RID p_map, real_t p_height) = 0;
@@ -90,20 +78,13 @@ public:
 	virtual void map_set_use_edge_connections(RID p_map, bool p_enabled) = 0;
 	virtual bool map_get_use_edge_connections(RID p_map) const = 0;
 
-	/// Set the map edge connection margin used to weld the compatible region edges.
 	virtual void map_set_edge_connection_margin(RID p_map, real_t p_connection_margin) = 0;
-
-	/// Returns the edge connection margin of this map.
 	virtual real_t map_get_edge_connection_margin(RID p_map) const = 0;
 
-	/// Set the map link connection radius used to attach links to the nav mesh.
 	virtual void map_set_link_connection_radius(RID p_map, real_t p_connection_radius) = 0;
-
-	/// Returns the link connection radius of this map.
 	virtual real_t map_get_link_connection_radius(RID p_map) const = 0;
 
-	/// Returns the navigation path to reach the destination from the origin.
-	virtual Vector<Vector3> map_get_path(RID p_map, Vector3 p_origin, Vector3 p_destination, bool p_optimize, uint32_t p_navigation_layers = 1) const = 0;
+	virtual Vector<Vector3> map_get_path(RID p_map, Vector3 p_origin, Vector3 p_destination, bool p_optimize, uint32_t p_navigation_layers = 1) = 0;
 
 	virtual Vector3 map_get_closest_point_to_segment(RID p_map, const Vector3 &p_from, const Vector3 &p_to, const bool p_use_collision = false) const = 0;
 	virtual Vector3 map_get_closest_point(RID p_map, const Vector3 &p_point) const = 0;
@@ -118,10 +99,18 @@ public:
 	virtual void map_force_update(RID p_map) = 0;
 	virtual uint32_t map_get_iteration_id(RID p_map) const = 0;
 
+	virtual void map_set_use_async_iterations(RID p_map, bool p_enabled) = 0;
+	virtual bool map_get_use_async_iterations(RID p_map) const = 0;
+
 	virtual Vector3 map_get_random_point(RID p_map, uint32_t p_navigation_layers, bool p_uniformly) const = 0;
 
-	/// Creates a new region.
+	/* REGION API */
+
 	virtual RID region_create() = 0;
+	virtual uint32_t region_get_iteration_id(RID p_region) const = 0;
+
+	virtual void region_set_use_async_iterations(RID p_region, bool p_enabled) = 0;
+	virtual bool region_get_use_async_iterations(RID p_region) const = 0;
 
 	virtual void region_set_enabled(RID p_region, bool p_enabled) = 0;
 	virtual bool region_get_enabled(RID p_region) const = 0;
@@ -129,89 +118,79 @@ public:
 	virtual void region_set_use_edge_connections(RID p_region, bool p_enabled) = 0;
 	virtual bool region_get_use_edge_connections(RID p_region) const = 0;
 
-	/// Set the enter_cost of a region
 	virtual void region_set_enter_cost(RID p_region, real_t p_enter_cost) = 0;
 	virtual real_t region_get_enter_cost(RID p_region) const = 0;
 
-	/// Set the travel_cost of a region
 	virtual void region_set_travel_cost(RID p_region, real_t p_travel_cost) = 0;
 	virtual real_t region_get_travel_cost(RID p_region) const = 0;
 
-	/// Set the node which manages this region.
 	virtual void region_set_owner_id(RID p_region, ObjectID p_owner_id) = 0;
 	virtual ObjectID region_get_owner_id(RID p_region) const = 0;
 
 	virtual bool region_owns_point(RID p_region, const Vector3 &p_point) const = 0;
 
-	/// Set the map of this region.
 	virtual void region_set_map(RID p_region, RID p_map) = 0;
 	virtual RID region_get_map(RID p_region) const = 0;
 
-	/// Set the region's layers
 	virtual void region_set_navigation_layers(RID p_region, uint32_t p_navigation_layers) = 0;
 	virtual uint32_t region_get_navigation_layers(RID p_region) const = 0;
 
-	/// Set the global transformation of this region.
 	virtual void region_set_transform(RID p_region, Transform3D p_transform) = 0;
 	virtual Transform3D region_get_transform(RID p_region) const = 0;
 
-	/// Set the navigation mesh of this region.
 	virtual void region_set_navigation_mesh(RID p_region, Ref<NavigationMesh> p_navigation_mesh) = 0;
 
 #ifndef DISABLE_DEPRECATED
-	/// Bake the navigation mesh.
 	virtual void region_bake_navigation_mesh(Ref<NavigationMesh> p_navigation_mesh, Node *p_root_node) = 0;
 #endif // DISABLE_DEPRECATED
 
-	/// Get a list of a region's connection to other regions.
 	virtual int region_get_connections_count(RID p_region) const = 0;
 	virtual Vector3 region_get_connection_pathway_start(RID p_region, int p_connection_id) const = 0;
 	virtual Vector3 region_get_connection_pathway_end(RID p_region, int p_connection_id) const = 0;
 
+	virtual Vector3 region_get_closest_point_to_segment(RID p_region, const Vector3 &p_from, const Vector3 &p_to, bool p_use_collision = false) const = 0;
+	virtual Vector3 region_get_closest_point(RID p_region, const Vector3 &p_point) const = 0;
+	virtual Vector3 region_get_closest_point_normal(RID p_region, const Vector3 &p_point) const = 0;
 	virtual Vector3 region_get_random_point(RID p_region, uint32_t p_navigation_layers, bool p_uniformly) const = 0;
 
-	/// Creates a new link between positions in the nav map.
+	virtual AABB region_get_bounds(RID p_region) const = 0;
+
+	/* LINK API */
+
 	virtual RID link_create() = 0;
+	virtual uint32_t link_get_iteration_id(RID p_link) const = 0;
 
-	/// Set the map of this link.
 	virtual void link_set_map(RID p_link, RID p_map) = 0;
 	virtual RID link_get_map(RID p_link) const = 0;
 
 	virtual void link_set_enabled(RID p_link, bool p_enabled) = 0;
 	virtual bool link_get_enabled(RID p_link) const = 0;
 
-	/// Set whether this link travels in both directions.
 	virtual void link_set_bidirectional(RID p_link, bool p_bidirectional) = 0;
 	virtual bool link_is_bidirectional(RID p_link) const = 0;
 
-	/// Set the link's layers.
 	virtual void link_set_navigation_layers(RID p_link, uint32_t p_navigation_layers) = 0;
 	virtual uint32_t link_get_navigation_layers(RID p_link) const = 0;
 
-	/// Set the start position of the link.
 	virtual void link_set_start_position(RID p_link, Vector3 p_position) = 0;
 	virtual Vector3 link_get_start_position(RID p_link) const = 0;
 
-	/// Set the end position of the link.
 	virtual void link_set_end_position(RID p_link, Vector3 p_position) = 0;
 	virtual Vector3 link_get_end_position(RID p_link) const = 0;
 
-	/// Set the enter cost of the link.
 	virtual void link_set_enter_cost(RID p_link, real_t p_enter_cost) = 0;
 	virtual real_t link_get_enter_cost(RID p_link) const = 0;
 
-	/// Set the travel cost of the link.
 	virtual void link_set_travel_cost(RID p_link, real_t p_travel_cost) = 0;
 	virtual real_t link_get_travel_cost(RID p_link) const = 0;
 
-	/// Set the node which manages this link.
 	virtual void link_set_owner_id(RID p_link, ObjectID p_owner_id) = 0;
 	virtual ObjectID link_get_owner_id(RID p_link) const = 0;
 
-	/// Creates the agent.
+	/* AGENT API */
+
 	virtual RID agent_create() = 0;
 
-	/// Put the agent in the map.
 	virtual void agent_set_map(RID p_agent, RID p_map) = 0;
 	virtual RID agent_get_map(RID p_agent) const = 0;
 
@@ -224,66 +203,37 @@ public:
 	virtual void agent_set_use_3d_avoidance(RID p_agent, bool p_enabled) = 0;
 	virtual bool agent_get_use_3d_avoidance(RID p_agent) const = 0;
 
-	/// The maximum distance (center point to
-	/// center point) to other agents this agent
-	/// takes into account in the navigation. The
-	/// larger this number, the longer the running
-	/// time of the simulation. If the number is too
-	/// low, the simulation will not be safe.
-	/// Must be non-negative.
 	virtual void agent_set_neighbor_distance(RID p_agent, real_t p_distance) = 0;
 	virtual real_t agent_get_neighbor_distance(RID p_agent) const = 0;
 
-	/// The maximum number of other agents this
-	/// agent takes into account in the navigation.
-	/// The larger this number, the longer the
-	/// running time of the simulation. If the
-	/// number is too low, the simulation will not
-	/// be safe.
 	virtual void agent_set_max_neighbors(RID p_agent, int p_count) = 0;
 	virtual int agent_get_max_neighbors(RID p_agent) const = 0;
 
-	// Sets the minimum amount of time in seconds that an agent's
-	// must be able to stay on the calculated velocity while still avoiding collisions with agent's
-	// if this value is set to high an agent will often fall back to using a very low velocity just to be safe
 	virtual void agent_set_time_horizon_agents(RID p_agent, real_t p_time_horizon) = 0;
 	virtual real_t agent_get_time_horizon_agents(RID p_agent) const = 0;
 
-	/// Sets the minimum amount of time in seconds that an agent's
-	// must be able to stay on the calculated velocity while still avoiding collisions with obstacle's
-	// if this value is set to high an agent will often fall back to using a very low velocity just to be safe
 	virtual void agent_set_time_horizon_obstacles(RID p_agent, real_t p_time_horizon) = 0;
 	virtual real_t agent_get_time_horizon_obstacles(RID p_agent) const = 0;
 
-	/// The radius of this agent.
-	/// Must be non-negative.
 	virtual void agent_set_radius(RID p_agent, real_t p_radius) = 0;
 	virtual real_t agent_get_radius(RID p_agent) const = 0;
 
 	virtual void agent_set_height(RID p_agent, real_t p_height) = 0;
 	virtual real_t agent_get_height(RID p_agent) const = 0;
 
-	/// The maximum speed of this agent.
-	/// Must be non-negative.
 	virtual void agent_set_max_speed(RID p_agent, real_t p_max_speed) = 0;
 	virtual real_t agent_get_max_speed(RID p_agent) const = 0;
 
-	/// forces and agent velocity change in the avoidance simulation, adds simulation instability if done recklessly
 	virtual void agent_set_velocity_forced(RID p_agent, Vector3 p_velocity) = 0;
 
-	/// The wanted velocity for the agent as a "suggestion" to the avoidance simulation.
-	/// The simulation will try to fulfill this velocity wish if possible but may change the velocity depending on other agent's and obstacles'.
 	virtual void agent_set_velocity(RID p_agent, Vector3 p_velocity) = 0;
 	virtual Vector3 agent_get_velocity(RID p_agent) const = 0;
 
-	/// Position of the agent in world space.
 	virtual void agent_set_position(RID p_agent, Vector3 p_position) = 0;
 	virtual Vector3 agent_get_position(RID p_agent) const = 0;
 
-	/// Returns true if the map got changed the previous frame.
 	virtual bool agent_is_map_changed(RID p_agent) const = 0;
 
-	/// Callback called at the end of the RVO process
 	virtual void agent_set_avoidance_callback(RID p_agent, Callable p_callback) = 0;
 	virtual bool agent_has_avoidance_callback(RID p_agent) const = 0;
 
@@ -296,7 +246,8 @@ public:
 	virtual void agent_set_avoidance_priority(RID p_agent, real_t p_priority) = 0;
 	virtual real_t agent_get_avoidance_priority(RID p_agent) const = 0;
 
-	/// Creates the obstacle.
+	/* OBSTACLE API */
+
 	virtual RID obstacle_create() = 0;
 
 	virtual void obstacle_set_map(RID p_obstacle, RID p_map) = 0;
@@ -324,41 +275,46 @@ public:
 	virtual void obstacle_set_avoidance_layers(RID p_obstacle, uint32_t p_layers) = 0;
 	virtual uint32_t obstacle_get_avoidance_layers(RID p_obstacle) const = 0;
 
-	/// Destroy the `RID`
-	virtual void free(RID p_object) = 0;
+	/* QUERY API */
 
-	/// Control activation of this server.
-	virtual void set_active(bool p_active) = 0;
+	virtual void query_path(const Ref<NavigationPathQueryParameters3D> &p_query_parameters, Ref<NavigationPathQueryResult3D> p_query_result, const Callable &p_callback = Callable()) = 0;
 
-	/// Process the collision avoidance agents.
-	/// The result of this process is needed by the physics server,
-	/// so this must be called in the main thread.
-	/// Note: This function is not thread safe.
-	virtual void process(real_t delta_time) = 0;
-	virtual void init() = 0;
-	virtual void sync() = 0;
-	virtual void finish() = 0;
-
-	/// Returns a customized navigation path using a query parameters object
-	virtual void query_path(const Ref<NavigationPathQueryParameters3D> &p_query_parameters, Ref<NavigationPathQueryResult3D> p_query_result) const;
-
-	virtual NavigationUtilities::PathQueryResult _query_path(const NavigationUtilities::PathQueryParameters &p_parameters) const = 0;
+	/* NAVMESH BAKE API */
 
 #ifndef _3D_DISABLED
 	virtual void parse_source_geometry_data(const Ref<NavigationMesh> &p_navigation_mesh, const Ref<NavigationMeshSourceGeometryData3D> &p_source_geometry_data, Node *p_root_node, const Callable &p_callback = Callable()) = 0;
 	virtual void bake_from_source_geometry_data(const Ref<NavigationMesh> &p_navigation_mesh, const Ref<NavigationMeshSourceGeometryData3D> &p_source_geometry_data, const Callable &p_callback = Callable()) = 0;
 	virtual void bake_from_source_geometry_data_async(const Ref<NavigationMesh> &p_navigation_mesh, const Ref<NavigationMeshSourceGeometryData3D> &p_source_geometry_data, const Callable &p_callback = Callable()) = 0;
 	virtual bool is_baking_navigation_mesh(Ref<NavigationMesh> p_navigation_mesh) const = 0;
+	virtual String get_baking_navigation_mesh_state_msg(Ref<NavigationMesh> p_navigation_mesh) const = 0;
 #endif // _3D_DISABLED
 
+protected:
+	static RWLock geometry_parser_rwlock;
+	static RID_Owner<NavMeshGeometryParser3D> geometry_parser_owner;
+	static LocalVector<NavMeshGeometryParser3D *> generator_parsers;
+
+public:
 	virtual RID source_geometry_parser_create() = 0;
 	virtual void source_geometry_parser_set_callback(RID p_parser, const Callable &p_callback) = 0;
 
 	virtual Vector<Vector3> simplify_path(const Vector<Vector3> &p_path, real_t p_epsilon) = 0;
 
+	/* SERVER API */
+
+	virtual void set_active(bool p_active) = 0;
+	virtual void process(double p_delta_time) = 0;
+	virtual void physics_process(double p_delta_time) = 0;
+	virtual void init() = 0;
+	virtual void sync() = 0;
+	virtual void finish() = 0;
+	virtual void free(RID p_object) = 0;
+
 	NavigationServer3D();
 	~NavigationServer3D() override;
 
+	/* DEBUG API */
+
 	enum ProcessInfo {
 		INFO_ACTIVE_MAPS,
 		INFO_REGION_COUNT,
@@ -369,6 +325,7 @@ public:
 		INFO_EDGE_MERGE_COUNT,
 		INFO_EDGE_CONNECTION_COUNT,
 		INFO_EDGE_FREE_COUNT,
+		INFO_OBSTACLE_COUNT,
 	};
 
 	virtual int get_process_info(ProcessInfo p_info) const = 0;
@@ -376,6 +333,13 @@ public:
 	void set_debug_enabled(bool p_enabled);
 	bool get_debug_enabled() const;
 
+protected:
+#ifndef DISABLE_DEPRECATED
+	Vector<Vector3> _map_get_path_bind_compat_100129(RID p_map, Vector3 p_origin, Vector3 p_destination, bool p_optimize, uint32_t p_navigation_layers = 1) const;
+	void _query_path_bind_compat_100129(const Ref<NavigationPathQueryParameters3D> &p_query_parameters, Ref<NavigationPathQueryResult3D> p_query_result) const;
+	static void _bind_compatibility_methods();
+#endif
+
 private:
 	bool debug_enabled = false;
 
@@ -559,8 +523,9 @@ class NavigationServer3DManager {
 public:
 	static void set_default_server(NavigationServer3DCallback p_callback);
 	static NavigationServer3D *new_default_server();
+
+	static void initialize_server();
+	static void finalize_server();
 };
 
 VARIANT_ENUM_CAST(NavigationServer3D::ProcessInfo);
-
-#endif // NAVIGATION_SERVER_3D_H
